import logging
import re
from base64 import b64decode
from datetime import datetime
from multiprocessing.sharedctypes import Value
from urllib.parse import parse_qsl, urlparse

from kibera.core import db
from kibera.core.metrics import key_requests_total

logger = logging.getLogger(__name__)


class KeyParseError(Exception):
    pass


class KeyExistsError(Exception):
    pass


def get_random_key():
    key_requests_total.inc()

    keys = db.keys.find({}, {"_id": 0}).sort("serveCount", 1).limit(1)
    keys = list(keys)
    if len(keys) == 0:
        return None
    key = keys[0]
    db.keys.update_one(
        {"host": key["host"], "port": key["port"]}, {"$inc": {"serveCount": 1}}
    )
    return key


def get_key(query):
    return db.keys.find_one(query, {"_id": 0})


def list_keys(query):
    return db.keys.find(query, {"_id": 0})


def create_key(user, key_data):
    host, port = key_data["host"], key_data["port"]
    if host is None or port is None:
        raise KeyParseError("Key must have 'host' and 'port'")

    now = datetime.utcnow().replace(microsecond=0)

    key = get_key({"host": host, "port": port})
    if key is not None:
        raise KeyExistsError("Key with host and port already exists")

    db.keys.insert_one(
        {**key_data, "createdAt": now, "createdBy": user, "serveCount": 0},
    )

    logger.info(f"Created key {host}:{port} by {user}")


def update_key(user, key_data):
    host, port = key_data["host"], key_data["port"]
    if host is None or port is None:
        raise KeyParseError("Need to specify 'host' and 'port' when updating key")

    now = datetime.utcnow().replace(microsecond=0)
    result = db.keys.update_one(
        {"host": host, "port": port},
        {"$set": {**key_data, "updatedAt": now, "updatedBy": user}},
    )

    if result.matched_count == 0:
        logger.info(f"Attempted updated for non-existing key {host}:{port} by {user}")
        return False

    logger.info(f"Updated key {host}:{port} by {user}")
    return True


def delete_key(user, key_data):
    host, port = key_data["host"], key_data["port"]
    if host is None or port is None:
        raise KeyParseError("Need to specify 'host' and 'port' when deleting key")

    db.keys.delete_one({"host": host, "port": port})

    logger.info(f"Deleted key {host}:{port} by {user}")


def parse_key_url(url):
    if "://" not in url:
        url = "ss://" + url

    key = {}
    p = urlparse(url)

    if p.scheme != "ss":
        raise KeyParseError(f"Unsupported key scheme '{p.scheme}'")

    m = re.match("(?:([^@]+)[@])?([^:]+)(?:[:]([0-9]+))?", p.netloc)
    if m is None:
        raise KeyParseError(f"Invalid key '{url}'")

    ciphersecret, host, port = m.groups()
    if host is not None:
        key["host"] = host
    if port is not None:
        key["port"] = int(port)

    if ciphersecret is not None:
        if ":" not in ciphersecret:
            ciphersecret = b64decode(ciphersecret + "===").decode("utf-8")
        if ":" not in ciphersecret:
            raise KeyParseError(f"Invalid cipher and secret in key '{url}'")
        key["cipher"], key["secret"] = ciphersecret.split(":")

    if p.fragment:
        key["name"] = p.fragment

    if p.query:
        key["param"] = (dict(parse_qsl(p.query)),)

    return key


def print_key(key):
    from rich import print_json

    print_json(data=key, sort_keys=True, default=str)



import logging
import os
import re
import signal
import sys
import threading

import click
import portforward

from kibera import keys, users
from kibera.core import db
from kibera.core.metrics import start_metrics_server

logger = logging.getLogger(__name__)


log_colors = (("uvicorn.*", 242),)


@click.group()
@click.option(
    "--log-level",
    default="info",
    type=click.Choice(
        ["debug", "info", "warning", "error", "critical"], case_sensitive=False
    ),
)
@click.option("--log-file", required=False, default=None)
@click.option("--redis", required=False, default="redis://127.0.0.1:6379")
@click.option("--mongo", required=False, default="mongodb://127.0.0.1:27017")
def cli(log_level, log_file, redis, mongo):
    from kibera.core.logging import init_logging

    init_logging(logging.getLevelName(log_level.upper()), log_file, colors=log_colors)
    db.set_url(mongo)


@cli.command()
@click.option("-p", "--port", default=8080)
@click.option("--metrics", required=False, default=9100)
def server(port, metrics):
    from waitress import serve

    from kibera.api import app

    def handler(signum, frame):
        print("Exiting...")
        sys.exit(0)

    signal.signal(signal.SIGINT, handler)
    signal.signal(signal.SIGTERM, handler)

    start_metrics_server(metrics_port=metrics)
    serve(app, listen=f"*:{port}")


@cli.command()
@click.option(
    "-k",
    "--kubeconfig",
    default=os.getenv("KUBECONFIG", os.path.expanduser("~/.kube/config")),
)
@click.option("--namespace", default="kibera")
@click.option("--pod", default="mongodb-0")
@click.option("--src-port", default=27017)
@click.option("--dst-port", default=27017)
def connect(kubeconfig, namespace, pod, src_port, dst_port):
    if not os.path.exists(kubeconfig):
        print(f"Kubeconfig not found at '{kubeconfig}'")
        sys.exit(1)
    print("Using kubeconfig %s", kubeconfig)

    ctx = portforward.forward(
        namespace,
        pod,
        src_port,
        dst_port,
        config_path=kubeconfig,
        log_level=portforward.LogLevel.ERROR,
    )

    lock = threading.Lock()

    def handler(signum, frame):
        lock.release
        print("Exiting...")
        sys.exit(0)

    lock.acquire()
    with ctx:
        print(f"Forwarding port {src_port} to kube {namespace}/{pod}:{dst_port}")
        print("Ctrl-C to exit")
        signal.signal(signal.SIGINT, handler)
        signal.signal(signal.SIGTERM, handler)
        lock.acquire()


@cli.group()
def add():
    pass


@cli.group()
def update():
    pass


@cli.group()
def delete():
    pass


@cli.group()
def get():
    pass


@add.command(name="key")
@click.argument("data", nargs=-1)
def add_key(data):
    key = parse_key_arg(data)
    keys.create_key("root", key)
    keys.print_key(keys.get_key(key))


@update.command(name="key")
@click.argument("data", nargs=-1)
def update_key(data):
    key = parse_key_arg(data)
    keys.update_key("root", key)
    keys.print_key(keys.get_key(key))


@get.command(name="key")
@click.argument("query", nargs=-1)
def get_key(query):
    data = parse_key_arg(query)
    key = keys.get_key(data)
    if key is None:
        print("Key not found")
        return
    keys.print_key(key)


@get.command(name="keys")
@click.argument("query", nargs=-1)
def get_keys(query):
    data = parse_key_arg(query)
    keylist = keys.list_keys(data)
    keys.print_key(list(keylist))


@delete.command(name="key")
@click.argument("query", nargs=-1)
def delete_key(query):
    data = parse_key_arg(query)
    keys.delete_key("root", data)


@add.command(name="user")
@click.argument("name")
def add_user(name):
    users.create(name)
    users.print_user(users.get(name))


@get.command(name="user")
@click.argument("name")
def get_user(name):
    user = users.get(name)
    if user is None:
        print("User not found")
        return
    users.print_user(user)


@get.command(name="users")
def get_users():
    users.print_user(list(users.list()))


@cli.command()
def shell():
    from kibera.core.shell import start_shell

    start_shell(db=db)


def parse_key_arg(key_data):
    qre = re.compile("([a-zA-Z0-9_-]+)=([a-zA-Z0-9_-]*)?")

    key = {}

    for s in key_data:
        m = qre.match(s)
        if m is not None:
            k, v = m.groups()
            key[k] = v
        else:
            key.update(keys.parse_key_url(s))

    if key.get("port") is not None:
        key["port"] = int(key["port"])

    return key


def main():
    cli(auto_envvar_prefix="KIBERA")



import functools
import logging
import re
import secrets
import string
from base64 import b64decode
from datetime import datetime
from multiprocessing.sharedctypes import Value
from urllib.parse import parse_qsl, urlparse

from flask import Flask, request
from werkzeug.exceptions import Unauthorized

from kibera.core import db

logger = logging.getLogger(__name__)

api_key_space = alphabet = string.ascii_letters + string.digits


class UserError(Exception):
    pass


def authenticate(func):
    @functools.wraps(func)
    def decorator(*args, **kwargs):
        api_key = None
        if request.data:
            api_key = request.json.get("api_key", None)
        if api_key is None:
            api_key = request.headers.get("api-key", None)
        if api_key is None:
            raise Unauthorized("Please provide an API key")
        user = db.users.find_one({"apiKey": api_key})
        if user is None:
            logger.warning("Auth attempted with invalid key: %s", api_key)
            raise Unauthorized("API key is invalid")
        logger.debug("User auth accepted: %s", user["name"])
        return func(user["name"], *args, **kwargs)

    return decorator


def create(name):
    user = db.users.find_one({"name": name})
    if user is not None:
        raise UserError("User already exists")

    user = {
        "name": name,
        "apiKey": "".join(secrets.choice(api_key_space) for i in range(32)),
    }
    db.users.insert_one(user)
    return user


def get(name):
    return db.users.find_one({"name": name}, {"_id": 0})


def list():
    return db.users.find({}, {"_id": 0})


def print_user(user):
    from rich import print_json

    print_json(data=user, sort_keys=True, default=str)



from prometheus_client import Counter, Histogram, start_http_server


def start_metrics_server(metrics_port):
    if metrics_port:
        start_http_server(metrics_port)


key_requests_total = Counter(
    "key_requests_total", "Number of requests made for fetching keys"
)

# To use: sample_metric.inc()



import logging
import re

from flask import Flask, g, request
from werkzeug.exceptions import BadRequest, HTTPException, NotFound, Unauthorized

from kibera import keys
from kibera.core import db
from kibera.users import authenticate

logger = logging.getLogger(__name__)

app = Flask(__name__)
app.url_map.strict_slashes = False


@app.before_request
def gather_request_data():
    g.method = request.method
    g.url = request.url


@app.after_request
def log_details(response):
    g.status = response.status_code
    logger.debug(f"{g.method} {g.url} {g.status}")
    return response


@app.errorhandler(HTTPException)
def handle_bad_request(e):
    return {"message": e.description}, 400


@app.route("/", methods=["GET"])
def get_random_key():
    key = keys.get_random_key()
    if key is None:
        return {}, NotFound
    return strip_key(key), 200


@app.route("/keys/", methods=["POST"])
@authenticate
def create_key(user):
    data = validate_key_data(request.json)
    try:
        keys.create_key(user, data)
    except keys.KeyExistsError:
        raise BadRequest("Key with given host and port already exists")
    return keys.get_key(data), 200


@app.route("/keys/", methods=["GET"])
@authenticate
def list_keys(user):
    data = validate_key_data(request.args, check_host_port=False)
    return list(keys.list_keys(data)), 200


@app.route("/key/<key>/", methods=["POST", "PUT"])
@app.route("/keys/<key>/", methods=["POST", "PUT"])
@authenticate
def update_key(user, key):
    data = keys.parse_key_url(key)
    data.update(request.json)
    data = validate_key_data(data)
    if keys.update_key(user, data):
        return keys.get_key(data), 200
    raise NotFound("Key not found")


@app.route("/key/<key>/", methods=["DELETE"])
@app.route("/keys/<key>/", methods=["DELETE"])
@authenticate
def delete_key(user, key):
    data = keys.parse_key_url(key)
    data = validate_key_data(data)
    keys.delete_key(user, data)
    return {"message": "Key deleted"}, 200


@app.route("/key/<key>/", methods=["GET"])
@app.route("/keys/<key>/", methods=["GET"])
@authenticate
def get_key(user, key):
    data = keys.parse_key_url(key)
    data.update(request.args)
    data = validate_key_data(data)
    k = keys.get_key(data)
    if k is None:
        raise NotFound(f"Key not found")
    return k, 200


def validate_key_data(data, check_host_port=True):
    if check_host_port:
        for k in ["port", "host"]:
            if not data.get(k):
                raise BadRequest(f"Missing field '{k}' in key data")
    if "port" in data:
        if type(data["port"]) is str and not data["port"].isnumeric():
            raise BadRequest(f"Invalid port {data['port']}")
        data["port"] = int(data["port"])
    return data


def strip_key(key):
    return {k: key.get(k) for k in ["host", "port", "cipher", "secret", "name"]}



import logging
import os
import random
import re
import threading
from urllib.parse import urlparse

import portforward
import pymongo

logger = logging.getLogger(__name__)

mongo_db_name = "kibera"


class DBError(Exception):
    pass


class DBProxy:
    _db = None
    _mongo_url = None
    _forward_port = 27018 + random.randint(1, 1000)
    _forward_ctx = None

    def _connect(self):
        if not self._mongo_url:
            raise ValueError("Mongo URL has not been set")

        logger.debug("Connecting to mongodb at %s", self._mongo_url)
        client = pymongo.MongoClient(self._mongo_url)
        self._db = client[mongo_db_name]

    def set_url(self, u):
        self._mongo_url = u

    def __getattr__(self, key):
        if self._db is None:
            self._connect()
        return getattr(self._db, key)


db = DBProxy()



import fnmatch
import logging

BLACK = 0
RED = 196
GREEN = 118
YELLOW = 229
BLUE = 75
MAGENTA = 206
CYAN = 45
WHITE = 255

RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[38;5;%dm"
BOLD_SEQ = "\033[1m"

COLORS = {
    "WARNING": YELLOW,
    "INFO": WHITE,
    "DEBUG": BLUE,
    "CRITICAL": YELLOW,
    "ERROR": RED,
}

# Taken from sentry lib
COMMON_RECORD_ATTRS = frozenset(
    (
        "args",
        "created",
        "exc_info",
        "exc_text",
        "filename",
        "funcName",
        "levelname",
        "levelno",
        "linenno",
        "lineno",
        "message",
        "module",
        "msecs",
        "msg",
        "name",
        "pathname",
        "process",
        "processName",
        "relativeCreated",
        "stack",
        "tags",
        "thread",
        "threadName",
        "stack_info",
        "asctime",
        "color_message",
    )
)


class CustomFormatter(logging.Formatter):
    def __init__(self, fmt, colors=[], colored_levels=True):
        logging.Formatter.__init__(self, fmt)
        self.colors = colors
        self.colored_levels = colored_levels
        self.cache = {}

    def formatMessage(self, record):
        msg = super().formatMessage(record)
        extras = " ".join(
            [
                f"{k}:{record.__dict__[k]}"
                for k in record.__dict__
                if k not in COMMON_RECORD_ATTRS
            ]
        )
        if not extras:
            return msg
        return f"{msg} | {extras}"

    def format(self, record: logging.LogRecord):
        levelname = record.levelname
        if self.colored_levels and levelname in COLORS:
            levelname_color = COLOR_SEQ % COLORS[levelname] + levelname + RESET_SEQ
            record.levelname = levelname_color

        # record.msg = record.name
        color = self.get_color(record.name)

        if color is not None:
            record.msg = COLOR_SEQ % color + record.msg + RESET_SEQ
            record.name = COLOR_SEQ % color + record.name + RESET_SEQ
        return super().format(record)

    def get_color(self, name):
        if name not in self.cache:
            color = None
            for match_with, c in self.colors:
                if fnmatch.fnmatch(name, match_with):
                    color = c
                    break
            self.cache[name] = color
        return self.cache[name]


def init_logging(loglevel=logging.INFO, logfile=None, colors=[], **kwargs):

    fmt = "%(asctime)s - %(levelname)s - %(name)s - %(message)s"

    if type(loglevel) is str:
        loglevel = logging.getLevelName(loglevel.upper())

    if logfile:
        handler = logging.FileHandler(logfile)
    else:
        handler = logging.StreamHandler()
    handler.setFormatter(CustomFormatter(fmt, colors))
    handler.setLevel(loglevel)

    logging.basicConfig(
        level=loglevel,
        handlers=[handler],
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kibera', 'kibera.core']

package_data = \
{'': ['*']}

install_requires = \
['flask>=2.2.2,<3.0.0',
 'portforward>=0.3.0,<0.4.0',
 'prometheus-client>=0.15.0,<0.16.0',
 'pymongo>=4.3.2,<5.0.0',
 'redis>=3.5.3,<4.0.0',
 'rich>=12.6.0,<13.0.0',
 'tqdm>=4.60.0,<5.0.0',
 'waitress>=2.1.2,<3.0.0']

entry_points = \
{'console_scripts': ['kibera = kibera:main']}

setup_kwargs = {
    'name': 'kibera',
    'version': '0.1.1',
    'description': '',
    'long_description': 'None',
    'author': 'Hadi',
    'author_email': 'hadi.zolfaghaari@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
